
typedef Vec<int, 16> IndexList ;
typedef Vec<int, 16> TargetList ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `MessSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class MessSchedule {
public :
  void gather(void* locBase, void* srvBase) ;
  void scatter(void* locBase, void* srvBase) ;
  void combine(void* locBase, void* srvBase, void (*op)(void*, void*)) ;
protected :
  MessSchedule(const int _elementLen) ;
  ~MessSchedule() ;

  void sendReq(int offset, int dstId) { (srvIdxs [dstId]).add(offset) ; }
  void recvReq(int offset, int srcId) { (locIdxs [srcId]).add(offset) ; }

  void build() ;

  int elementLen ;

  int nProcs, locId ;

  IndexList* locIdxs ;
  IndexList* srvIdxs ;

  TargetList locList ;
  TargetList srvList ;

  char* locBuf ;
  char* srvBuf ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `SendServer' class.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class SendServer : public Server {
public :
  SendServer(const int numMes, int* remIds,
             char* _sendBuf, IndexList* _sendIdxs,
             void* _sendBase, int _elementLen) ;
  ~SendServer() ;

  void transition(MPI_Status* status) ;

  int active() const ;

  void initSend(const int remId) ;
private :
  int numMes ;
  int *remIds ;

  char* sendBuf ;
  IndexList* sendIdxs ;

  void* sendBase ;
  int elementLen ;

  int mes ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockMessSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockMessSchedule : protected GlobalEnv {
public :
  BlockMessSchedule(const int _elementLen, const int _rank) :
      locMessages(nProcs), srvMessages(nProcs), cpyNum(0),
      elementLen(_elementLen), rank(_rank) {}

  void sendReq(int offset, int* strs, int ctg,
               int* exts, int vol, const int dstId) {
    if(dstId == locId) {
      addBlock(srvCopies, offset, strs, ctg, elementLen * vol, exts, rank) ;
      cpyNum++ ;
    }
    else {
      const int hop = (dstId - locId + nProcs) % nProcs ;
      MessQueue* queue = srvMessages.getElement(hop) ;
      addMessageBlock(queue->messages, queue->blocks,
                      offset, strs, ctg, elementLen * vol, exts, rank) ;
      queue->remId = dstId ;
      queue->hop   = hop   ;
    }
  }

  void recvReq(int offset, int* strs, int ctg,
               int* exts, int vol, const int srcId) {
    if(srcId == locId)
      addBlock(locCopies, offset, strs, ctg, elementLen * vol, exts, rank) ;
    else {
      const int hop = (locId - srcId + nProcs) % nProcs ;
      MessQueue* queue = locMessages.getElement(hop) ;
      addMessageBlock(queue->messages, queue->blocks,
                      offset, strs, ctg, elementLen * vol, exts, rank) ;
      queue->remId = srcId ;
      queue->hop   = hop   ;
    }
  }

  void build() {
    locList = locMessages.toList() ;
    srvList = srvMessages.toList() ;
  }

  void gather(void* locBase, void* srvBase) {
    run(locList, srvList, &locCopies, &srvCopies, locBase, srvBase) ;
  }

  void scatter(void* locBase, void* srvBase) {
    run(srvList, locList, &srvCopies, &locCopies, srvBase, locBase) ;
  }

  void show() ;

protected :

  int elementLen, rank ;

  SparseArray<MessQueue> locMessages, srvMessages ;
  List<MessQueue> *locList, *srvList ;

  MultiBlockList locCopies, srvCopies ;
  int cpyNum ;

private :
  void run(List<MessQueue>* dstMessages, List<MessQueue>* srcMessages,
           MultiBlockList* dstCopies, MultiBlockList* srcCopies,
           void* dstBase, void* srcBase) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockSendServer' class.                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockSendServer : public Server {
public :
  BlockSendServer(List<MessQueue>* sendList, void* _sendBase, int _elementLen) ;

  void transition(MPI_Status* status) ;

  void* active() const ;

private :
  void initSend() ;

  List<MessQueue>* record ;
  int remId ;
  int numMes ;

  int mes ;
  Mess* message ;
  MultiBlock block ;

  void* sendBase ;
  int elementLen ;

  align_t sendBuf [(STATIC_BUFFER_SIZE + ALIGN_SZ - 1) / ALIGN_SZ] ;
} ;

// ----------------------------------------------------------------------
// Things to do
//
//    Lose base addresses from schedules?  Pass them in at schedule
//    execution time.

